#!/usr/bin/env python

from ..action import Action

class ConsoleTarget:
	def read (self, logger, relative):
		raise Exception ('can\'t read from console target')

	def send (self, logger, work, actions):
		for action in actions:
			if action.type == action.ADD:
				prefix = '((lime))+'
			elif action.type == action.DEL:
				prefix = '((blue))-'
			elif action.type != action.NOP:
				prefix = '((red))!'
			else:
				continue

			logger.info (prefix + '((reset)) ' + action.path)

		return True



#!/usr/bin/env python

from ..action import Action
from ..process import Process

class GitSource:
	def current (self, base_path):
		revision = Process (['git', 'rev-parse', '--quiet', '--verify', 'HEAD']) \
			.set_directory (base_path) \
			.execute ()

		if revision:
			return revision.out.decode ('utf-8').strip ()

		return None

	def diff (self, logger, base_path, work_path, rev_from, rev_to):
		# Parse and validate source and target revisions
		if rev_from is not None and rev_from != '':
			process = Process (['git', 'rev-parse', '--quiet', '--verify', rev_from])
		else:
			process = Process (['git', 'hash-object', '-t', 'tree', '/dev/null'])

		res_from = process \
			.set_directory (base_path) \
			.execute ()

		if not res_from:
			logger.warning ('Invalid source revision "{0}".'.format (rev_from))
			logger.debug (res_from.err.decode ('utf-8'))

			return None

		res_to = Process (['git', 'rev-parse', '--quiet', '--verify', rev_to]) \
			.set_directory (base_path) \
			.execute ()

		if not res_to:
			logger.warning ('Invalid target revision "{0}".'.format (rev_to))
			logger.debug (res_to.err.decode ('utf-8'))

			return None

		hash_from = res_from.out.decode ('utf-8').strip ()
		hash_to = res_to.out.decode ('utf-8').strip ()

		# Display update information
		if hash_from != hash_to:
			logger.info ('Update from revision ((fuchsia)){0}((reset)) to ((fuchsia)){1}((reset))...'.format (hash_from[0:8], hash_to[0:8]))
		else:
			logger.info ('Already at revision ((fuchsia)){0}((reset)).'.format (hash_from[0:8]))

			return []

		# Populate work directory from Git archive
		archive = Process (['git', 'archive', hash_to, '.']) \
			.set_directory (base_path) \
			.pipe (['tar', 'xC', work_path]) \
			.execute ()

		if not archive:
			logger.warning ('Couldn\'t export archive from Git.')
			logger.debug (archive.err.decode ('utf-8'))

			return None

		# Build actions from Git diff output
		diff = Process (['git', 'diff', '--name-status', '--relative', hash_from, hash_to]) \
			.set_directory (base_path) \
			.execute ()

		if not diff:
			logger.warning ('Couldn\'t get diff from Git.')
			logger.debug (diff.err.decode ('utf-8'))

			return None

		actions = []

		for line in diff.out.decode ('utf-8').splitlines ():
			(mode, path) = line.split ('\t', 1)

			if mode == 'A' or mode == 'M':
				actions.append (Action (path, Action.ADD))
			elif mode == 'D':
				actions.append (Action (path, Action.DEL))

		return actions



#!/usr/bin/env python

import codecs
import json
import os
import re

from .action import Action
from .process import Process

class DefinitionModifier:
	def __init__ (self, regex, filter, rename, modify, link):
		self.filter = filter
		self.link = link
		self.modify = modify
		self.regex = regex
		self.rename = rename

class Definition:
	def __init__ (self, file, ignores):
		modifiers = []

		# Read modifiers from configuration file if available...
		if file is not None:
			reader = codecs.getreader ('utf-8')
			config = json.load (reader (file))

			for modifier in config.get ('modifiers', []):
				modify = modifier.get ('modify', modifier.get ('adapt', None))
				filter = modifier.get ('filter', None)
				link = modifier.get ('link', None)
				rename = modifier.get ('rename', modifier.get ('name', None))

				modifiers.append (DefinitionModifier (re.compile (modifier['pattern']), filter, rename, modify, link))

			options = config.get ('options', {})
			source = config.get ('source', None)

		# ...or provide default set of modifiers otherwise
		else:
			options = {}
			source = None

		for ignore in ignores:
			modifiers.append (DefinitionModifier (re.compile ('^' + re.escape (ignore) + '$'), '', None, None, None))

		self.modifiers = modifiers
		self.options = options
		self.source = source

	def apply (self, logger, work, path, type, used):
		# Ensure we don't process a file already scanned
		path = os.path.normpath (path)

		if path in used:
			return ([], [])

		used.add (path)

		# Find modifier matching current file name if any
		name = os.path.basename (path)

		for modifier in self.modifiers:
			match = modifier.regex.search (name)

			if match is None:
				continue

			logger.debug ('File \'{0}\' matches \'{1}\'.'.format (path, modifier.regex.pattern))

			actions = []
			cancels = []

			# Apply renaming pattern if any
			if modifier.rename is not None:
				name = os.path.basename (re.sub ('\\\\([0-9]+)', lambda m: match.group (int (m.group (1))), modifier.rename))

				logger.debug ('File \'{0}\' renamed to \'{1}\'.'.format (path, name))

			path_new = os.path.normpath (os.path.join (os.path.dirname (path), name))

			if type == Action.ADD:
				# Apply link command if any
				if modifier.link is not None:
					out = self.run (work, path, modifier.link)

					if out is not None:
						for link in out.splitlines ():
							logger.debug ('File \'{0}\' is linked to file \'{1}\'.'.format (path, link))

							(actions_append, cancels_append) = self.apply (logger, work, link, type, used)

							actions.extend (actions_append)
							cancels.extend (cancels_append)
					else:
						logger.debug ('Command \'link\' on file \'{0}\' returned non-zero code.'.format (path))

				# Build output file using processing command if any
				if modifier.modify is not None:
					out = self.run (work, path, modifier.modify)

					if out is not None:
						with open (os.path.join (work, path_new), 'wb') as file:
							file.write (out)
					else:
						logger.debug ('Command \'modify\' on file \'{0}\' returned non-zero code.'.format (path))

						type = Action.ERR

					if path != path_new:
						cancels.append (path)

				# Otherwise, copy original to renamed
				elif path != path_new:
					shutil.copy (os.path.join (work, path), os.path.join (work, path_new))

					cancels.append (path)

			# Apply filtering command if any
			if modifier.filter is not None and (modifier.filter == '' or self.run (work, path, modifier.filter) is None):
				logger.debug ('File \'{0}\' filtered out.'.format (path))

				type = Action.NOP

			# Append action to list and return
			actions.append (Action (path_new, type))

			return (actions, cancels)

		# No modifier matched, return unmodified input
		return ([Action (path, type)], [])

	def run (self, work, path, command):
		result = Process (command.replace ('{}', path)) \
			.set_directory (work) \
			.set_shell (True) \
			.execute ()

		if not result:
			return None

		return result.out



#!/usr/bin/env python

import hashlib
import os

from ..action import Action
from .. import path

class HashSource:
	def __init__ (self, options):
		self.algorithm = options.get ('algorithm', 'md5')
		self.follow = options.get ('follow', True)

	def current (self, base_path):
		entries = {}

		for name in os.listdir (base_path):
			source = os.path.join (base_path, name)

			if not self.follow and os.path.islink (source):
				continue
			elif os.path.isdir (source):
				entry = self.current (source)
			elif os.path.isfile (source):
				entry = self.digest (source)
			else:
				continue

			entries[name] = entry

		return entries

	def diff (self, logger, base_path, work_path, rev_from, rev_to):
		return self.recurse (base_path, work_path, '.', rev_from or {}, rev_to or {})

	def digest (self, path):
		hash = hashlib.new (self.algorithm)

		with open (path, 'rb') as file:
			for chunk in iter (lambda: file.read (4096), b''):
				hash.update (chunk)

		return hash.hexdigest ()

	def recurse (self, base_path, work_path, parent, entries_from, entries_to):
		actions = []

		for name in set (entries_from.keys ()).union (entries_to.keys ()):
			entry_from = entries_from.get (name, None)
			entry_to = entries_to.get (name, None)
			source = os.path.join (parent, name)

			# Path was a directory on previous version
			if isinstance (entry_from, dict):
				# Path is still a directory => compare recursively
				if isinstance (entry_to, dict):
					actions.extend (self.recurse (base_path, work_path, source, entry_from, entry_to))

				# Path is no longer a directory
				else:
					# Path is now a file => add
					if entry_to is not None and path.duplicate (os.path.join (base_path, source), work_path, source):
						actions.append (Action (source, Action.ADD))

					# Recurse with no right hand side to delete contents
					actions.extend (self.recurse (base_path, work_path, source, entry_from, {}))

			# Path wasn't a directory on previous version but now is
			elif isinstance (entry_to, dict):
				# Path was a file => delete
				if entry_from is not None and path.duplicate (os.path.join (base_path, source), work_path, source):
					actions.append (Action (source, Action.DEL))

				# Recurse with no left hand side to add contents
				actions.extend (self.recurse (base_path, work_path, source, {}, entry_to))

			# Path wasn't and isn't a directory but changed
			elif entry_from != entry_to:
				# Path is now a file => add
				if entry_to is not None and path.duplicate (os.path.join (base_path, source), work_path, source):
					actions.append (Action (source, Action.ADD))

				# Path no longer exists => delete
				else:
					actions.append (Action (source, Action.DEL))

		return actions



#!/usr/bin/env python

import codecs
import json

class EnvironmentLocation:
	def __init__ (self, location):
		self.append_files = location.get ('append_files', [])
		self.cascades = dict (((path, isinstance (name, list) and name or [name]) for path, name in location.get ('cascades', location.get ('subsidiaries', {})).items ()))
		self.connection = location.get ('connection', None)
		self.local = location.get ('local', False)
		self.options = location.get ('options', {})
		self.remove_files = location.get ('remove_files', [])
		self.state = location.get ('state', '.creep.rev')

class Environment:
	def __init__ (self, file):
		# Load environments configuration from file
		try:
			reader = codecs.getreader ('utf-8')
			config = json.load (reader (file))

		except KeyError as key:
			raise ValueError ('missing property "{0}" in environments file'.format (key))

		self.locations = dict (((name, EnvironmentLocation (location)) for (name, location) in config.items ()))

	def get_location (self, name):
		return self.locations.get (name, None)



#!/usr/bin/env python

import json

class Revision:
	def __init__ (self, data):
		states = {}	

		if data != '':
			for (name, rev) in json.loads (data.decode ('utf-8')).items ():
				states[name] = rev

		self.states = states

	def get (self, name):
		return self.states.get (name, None)

	def serialize (self):
		return json.dumps (self.states, indent = 4, sort_keys = True)

	def set (self, name, data):
		self.states[name] = data



#!/usr/bin/env python

# Hack for Python 2 + 3 compatibility
try:
    from urllib.parse import unquote_plus
except ImportError:
    from urllib import unquote_plus

from . import path

import os
import re
import urllib

def create_source (source, options, directory):
	source = source or detect (directory)

	if source == 'delta' or source == 'hash':
		from .sources.hash import HashSource

		return HashSource (options)

	if source == 'git':
		from .sources.git import GitSource

		return GitSource ()

	# No known source type recognized
	return None

def create_target (logger, connection, options):
	# FIXME: should use urllib.parse [url-parse]
	match = re.match ('([+0-9A-Za-z]+)://(?:([^#/:@]+)(?::([^#/@]+))?@)?(?:([^#/:]+)(?::([0-9]+))?)?(?:/([^#]*))?', connection)

	if match is None:
		return None

	directory = match.group (6) or '.'
	host = match.group (4)
	password = match.group (3) is not None and unquote_plus (match.group (3)) or match.group (3)
	port = match.group (5) is not None and int (match.group (5)) or None
	scheme = match.group (1)
	user = match.group (2) is not None and unquote_plus (match.group (2)) or match.group (2)

	if scheme == 'file':
		if password is not None or port is not None or user is not None:
			logger.warn ('Connection string for "file" scheme shouldn\'t contain any port, user or password.')

		from .targets.file import FileTarget

		return FileTarget (directory)

	if scheme == 'ftp':
		from .targets.ftp import FTPTarget

		return FTPTarget (host, port, user, password, directory, options)

	if scheme == 'ssh':
		if password is not None:
			logger.warn ('Connection string for "ssh" scheme shouldn\'t contain any password.')

		from .targets.ssh import SSHTarget

		return SSHTarget (host, port, user, directory, options)

	# No known scheme recognized
	return None

def detect (directory):
	drive, tail = os.path.splitdrive (os.path.abspath (directory))
	names = path.explode (tail)

	# Detect '.git' file or directory in parent folders
	if any ((os.path.exists (drive + os.path.join (*(names[0:n] + ['.git']))) for n in range (len (names), 0, -1))):
		return 'git'

	# Fallback to hash source by default
	return 'hash'



#!/usr/bin/env python

from . import factory, path
from .action import Action
from .definition import Definition
from .environment import Environment
from .revision import Revision

import os
import shutil
import tempfile

def execute (logger, base_path, definition_path, environment_path, names, append_files, remove_files, rev_from, rev_to, yes):
	# Ensure base directory is valid
	if not os.path.isdir (base_path):
		logger.error ('Base directory "{0}" doesn\'t exist.'.format (base_path))

		return False

	# Load environment file or fail
	full_path = os.path.join (base_path, environment_path)

	if os.path.isfile (full_path):
		with open (full_path, 'rb') as file:
			environment = Environment (file)
	else:
		logger.error ('No environment file "{0}" found.'.format (full_path))

		return False

	# Load definition file or use default
	full_path = os.path.join (base_path, definition_path)

	if os.path.isfile (full_path):
		with open (full_path, 'rb') as file:
			definition = Definition (file, [environment_path, definition_path])
	else:
		definition = Definition (None, [environment_path, definition_path])

	# Expand location names
	if len (names) < 1:
		names.append ('default')
	elif len (names) == 1 and names[0] == '*':
		names = environment.locations.keys ()

	# Deploy to target locations
	ok = True

	for name in names:
		location = environment.get_location (name)

		if location is None:
			logger.error ('There is no location "{0}" in your environment file.'.format (name))

			continue

		if location.connection is not None:
			logger.info ('Deploying to location "{0}"...'.format (name))

			if not sync (logger, base_path, definition, location, name, append_files, remove_files, rev_from, rev_to, yes):
				ok = False

				continue

		for cascade_path, cascade_names in location.cascades.items ():
			full_path = os.path.join (base_path, cascade_path)

			logger.info ('Cascading to path "{0}"...'.format (full_path))
			logger.enter ()

			ok = execute (logger, full_path, definition_path, environment_path, cascade_names, [], [], None, None, yes) and ok

			logger.leave ()

	return ok

def prompt (logger, question):
	logger.info (question)

	while True:
		answer = input ()

		if answer == 'N' or answer == 'n':
			return False
		elif answer == 'Y' or answer == 'y':
			return True

		logger.warning ('Invalid answer')

def sync (logger, base_path, definition, location, name, append_files, remove_files, rev_from, rev_to, yes):
	# Build source repository reader from current directory
	source = factory.create_source (definition.source, definition.options, base_path)

	if source is None:
		logger.error ('Unknown source type in folder "{0}", try specifying "source" option in definition file.'.format (base_path))

		return False

	# Build target from location connection string
	target = factory.create_target (logger, location.connection, location.options)

	if target is None:
		logger.error ('Unsupported scheme in connection string "{0}".'.format (location.connection))

		return False

	# Read revision file
	if not location.local:
		data = target.read (logger, location.state)
	elif os.path.exists (os.path.join (base_path, location.state)):
		data = open (os.path.join (base_path, location.state), 'rb').read ()
	else:
		data = ''

	if data is None:
		logger.error ('Can\'t read revision file "{0}", check connection string and ensure parent directory exists.'.format (location.state))

		return False

	try:
		revision = Revision (data)
	except Exception as e:
		logger.error ('Can\'t parse revision from file "{0}": {1}.'.format (location.state, e))

		return False

	# Retrieve source and target revision
	if rev_from is None:
		rev_from = revision.get (name)

		if rev_from is None and not yes and not prompt (logger, 'No current revision found, maybe you\'re deploying for the first time. Initiate full deploy? [Y/N]'):
			return True

	if rev_to is None:
		rev_to = source.current (base_path)

		if rev_to is None:
			logger.error ('Can\'t find source version, please ensure your environment file is correctly defined.')

			return False

	revision.set (name, rev_to)

	# Prepare actions
	work_path = tempfile.mkdtemp ()

	try:
		# Append actions from revision diff
		source_actions = source.diff (logger, base_path, work_path, rev_from, rev_to)

		if source_actions is None:
			return False

		# Append actions for manually specified files
		manual_actions = []

		for append in location.append_files + append_files:
			full_path = os.path.join (base_path, append)

			if os.path.isdir (full_path):
				for (dirpath, dirnames, filenames) in os.walk (full_path):
					parent_path = os.path.relpath (dirpath, base_path)

					manual_actions.extend ((Action (os.path.join (parent_path, filename), Action.ADD) for filename in filenames))
			elif os.path.isfile (full_path):
				manual_actions.append (Action (append, Action.ADD))
			else:
				logger.warning ('Can\'t append missing file "{0}".'.format (append))

		for action in manual_actions:
			if not path.duplicate (os.path.join (base_path, action.path), work_path, action.path):
				logger.warning ('Can\'t copy file "{0}".'.format (action.path))

		for remove in location.remove_files + remove_files:
			full_path = os.path.join (base_path, remove)

			if os.path.isdir (full_path):
				for (dirpath, dirnames, filenames) in os.walk (full_path):
					parent_path = os.path.relpath (dirpath, base_path)

					manual_actions.extend ((Action (os.path.join (parent_path, filename), Action.DEL) for filename in filenames))
			else:
				manual_actions.append (Action (remove, Action.DEL))

		# Apply pre-processing modifiers on actions
		actions = []
		cancels = []
		used = set ()

		for command in source_actions + manual_actions:
			(actions_append, cancels_append) = definition.apply (logger, work_path, command.path, command.type, used)

			actions.extend (actions_append)
			cancels.extend (cancels_append)

		for cancel in cancels:
			path.remove (work_path, cancel)

		# Update current revision (remote mode)
		if rev_from != rev_to and not location.local:
			with open (os.path.join (work_path, location.state), 'wb') as file:
				file.write (revision.serialize ().encode ('utf-8'))

			actions.append (Action (location.state, Action.ADD))

		# Display processed actions using console target
		if len (actions) < 1:
			logger.info ('No deployment required.')

			return True

		from .targets.console import ConsoleTarget

		console = ConsoleTarget ()
		console.send (logger, work_path, actions)

		if not yes and not prompt (logger, 'Deploy? [Y/N]'):
			return True

		# Execute processed actions after ordering them by precedence
		actions.sort (key = lambda action: (action.order (), action.path))

		if not target.send (logger, work_path, actions):
			return False

		# Update current revision (local mode)
		if location.local:
			with open (os.path.join (base_path, location.state), 'wb') as file:
				file.write (revision.serialize ().encode ('utf-8'))

	finally:
		shutil.rmtree (work_path)

	logger.info ('Deployment done.')

	return True

# Hack for Python 2 + 3 compatibility
try:
	input = raw_input
except NameError:
	pass



#!/usr/bin/env python

import subprocess

class ProcessResult:
	def __init__ (self, code, out, err):
		self.code = code
		self.err = err
		self.out = out

	def __bool__ (self):
		return self.code == 0

	def __nonzero__ (self):
		return self.code == 0

class Process:
	def __init__ (self, args):
		self.args = args
		self.directory = None
		self.input = None
		self.parent = None
		self.shell = False
		self.stdin = None

	def _open (self, stdout):
		process = subprocess.Popen (self.args, cwd = self.directory, shell = self.shell, stderr = subprocess.PIPE, stdin = self.stdin, stdout = stdout)

		if self.parent is not None:
			self.parent._open (process.stdin)

		return process

	def execute (self):
		process = self._open (subprocess.PIPE)

		(out, err) = process.communicate (self.input)

		return ProcessResult (process.returncode, out, err)

	def pipe (self, args):
		next = Process (args)
		next.parent = self
		next.stdin = subprocess.PIPE

		return next

	def set_directory (self, directory):
		self.directory = directory

		return self

	def set_input (self, input):
		self.input = input
		self.stdin = subprocess.PIPE

		return self

	def set_shell (self, shell):
		self.shell = shell

		return self
